"""RTMP packet functions."""

from _exceptions import PacketError
from AMF.Deserializer import Deserializer
from AMF.InputStream import InputStream
from AMF.OutputStream import OutputStream
from AMF.Serializer import Serializer

import struct

TYPE_CHUNK_SIZE = 0x01
TYPE_BYTES_READ = 0x03
TYPE_PING       = 0x04
TYPE_AUDIO      = 0x08
TYPE_VIDEO      = 0x09
TYPE_METADATA   = 0x12
TYPE_FUNCTION   = 0x14

class RTMPPacket:
    """An RTMP packet.

    If packet.type == TYPE_AUDIO or TYPE_VIEDO, then the following fields are
    set:
    - packet.data contains raw audio or video data.

    If packet.type == TYPE_FUNCTION, then the following fields are set:
    - packet.function is the function name.
    - packet.id is the callback ID.
    - packet.args is a tuple of the arguments.

    """
    def __init__(self, channel):
        self.type      = channel.type
        self.timer     = channel.timer
        self.channel   = channel.number
        self.stream_id = channel.stream_id
        if channel.type == TYPE_FUNCTION:
            self.data     = deserialize(channel.data)
            self.function = self.data[0]
            self.id       = self.data[1]
            self.args     = tuple(self.data[2:])
        elif channel.type == TYPE_METADATA:
            self.data     = deserialize(channel.data)
            self.function = None
            self.id       = None
            self.args     = None
        else:
            self.data     = channel.data
            self.function = None
            self.id       = None
            self.args     = None

    def __repr__(self):
        more_info = ''
        if self.type == TYPE_CHUNK_SIZE:
            typestr = "Chunk Size"
        elif self.type == TYPE_BYTES_READ:
            typestr = 'Bytes Read'
        elif self.type == TYPE_PING:
            typestr = 'Ping'
        elif self.type == TYPE_VIDEO:
            typestr = 'Video'
        elif self.type == TYPE_AUDIO:
            typestr = 'Audio'
        elif self.type == TYPE_METADATA:
            typestr = 'Metadata'
        elif self.type == TYPE_FUNCTION:
            typestr = 'Function'
            more_info = ''.join((' (', self.function, ')'))
        else:
            typestr = hex(self.type)
        return ''.join(('<RTMPPacket type=', typestr, more_info, '>'))


class Channel:
    def __init__(self, number):
        self.number    = number
        # These are cleared after the message is received:
        self.remaining = None           # Bytes remaining in the message.
        self.data      = None           # Message body.
        # These are saved between messages, because the protocol may omit them
        # to save space if they are the same on the next message:
        self.length    = None           # Length of the message.
        self.type      = None           # Message type, e.g. TYPE_VIDEO.
        self.timer     = None           # Message timer.
        self.stream_id = None           # Stream ID.
        # The call() id is saved so we don't use it again later.
        self.call_id   = None

    def open(self, type, length, timer, stream_id):
        """Open this channel to prepare it for data.
        
        If either type or length == None, type and length information from the
        previous message will be used.

        """
        if type == None or length == None:
            if self.length == None or self.type == None:
                raise PacketError, "Length and type were omitted, but no "\
                                   "previous information is available"
            else:
                self.remaining = self.length
        else:
            self.remaining = length
            self.length    = length
            self.type      = type
        if timer == None:
            if self.timer == None:
                raise PacketError, "Timer was omitted, but no previous "\
                                   "information is available"
        else:
            self.timer = timer
        if stream_id == None:
            if self.stream_id == None:
                raise PacketError, "Stream ID was omitted, but no previous "\
                                   "information is available"
        else:
            self.stream_id = stream_id
        self.data = ""

    def close(self):
        """Close this channel, wiping out the data."""
        self.remaining = None
        self.data      = None

    def is_open(self):
        """Return True if this channel is currently open."""
        if self.data != None:
            return True
        else:
            return False


class PacketReader:
    """Read in raw RTMP protocol and emit RTMPPacket objects."""
    def __init__(self):
        # Remember these between messages:
        self.channels = {}              # Channel number -> channel instance.
        self.current_channel = None     # Channel # we're currently reading.
        self.bytes_left = None          # Bytes left on the current channel.
        self.incomplete_header = None   # Header we're currently reading.
        self.chunk_size = 128           # Bytes of data between headers.

    def parse(self, data):
        """Read the given RTMP data, and return a list of parsed packets."""
        result = []
        while len(data) != 0:
            if self.current_channel == None or self.bytes_left == 0:
                # New RTMP message.
                if self.incomplete_header == None:
                    data_offset = 0
                    self.incomplete_header  = data
                else:
                    data_offset = -len(self.incomplete_header)
                    self.incomplete_header += data
                channel_num = ord(self.incomplete_header[0]) & 0x3F
                header_type = ord(self.incomplete_header[0]) >> 6
                if   header_type == 0x00: header_length = 12
                elif header_type == 0x01: header_length = 8
                elif header_type == 0x02: header_length = 4
                else:                     header_length = 1
                if len(self.incomplete_header) < header_length:
                    # Not enough bytes for the full header.  Try again later.
                    data = ''
                    continue

                self.current_channel = channel_num
                if not channel_num in self.channels:
                    self.channels[channel_num] = Channel(channel_num)
                channel = self.channels[channel_num]

                header = self.incomplete_header[0 : header_length]
                data   = data[data_offset + header_length :]
                self.incomplete_header = None

                if header_length != 1:      # Pad 1st byte to make 32-bit.
                    timer = struct.unpack(">I", '\x00' + header[1:4])[0]
                else:
                    timer = None
                if header_length == 8 or header_length == 12: # Also pad.
                    length = struct.unpack(">I", '\x00' + header[4:7])[0]
                    type   = ord(header[7])
                else:
                    length = None
                    type   = None
                if header_length == 12:     # Note: LITTLE ENDIAN.
                    stream_id = struct.unpack("<I", header[8:12])[0]
                else:
                    stream_id = None

                if not channel.is_open():
                    channel.open(type, length, timer, stream_id)
                self.bytes_left = min(channel.remaining, self.chunk_size)
            else:
                channel = self.channels[self.current_channel]

            newdata = data[:self.bytes_left]
            data    = data[self.bytes_left:]
            channel.data      += newdata
            channel.remaining -= len(newdata)
            self.bytes_left   -= len(newdata)

            if channel.remaining == 0:
                # Finished a packet!
                newpacket = RTMPPacket(channel)
                if newpacket.type == TYPE_CHUNK_SIZE:
                    self.chunk_size = struct.unpack(">I", newpacket.data)[0]
                else: # Send all non-TYPE_CHUNK_SIZE packets to the client.
                    result.append(newpacket)
                if channel.type == TYPE_FUNCTION:
                    channel.call_id = int(newpacket.data[1])
                channel.close()
                self.current_channel = None
                self.bytes_left = None
        # end while
        return result


def serialize(*data):
    """Serialize a bunch of data into AMF format."""
    out = OutputStream()
    serializer = Serializer(out, None)
    for datum in data:
        serializer.writeData(datum, None)
    return out.flush()

def deserialize(data):
    """Deserialize an AMF block."""
    inp = InputStream(data)
    deserializer = Deserializer(inp, None)
    result = []
    while inp.current_byte < inp.content_length:
        result.append(deserializer.readSingleData())
    return tuple(result)
